unit GCAModule;

interface

uses
	Windows, SysUtils, Math,
	UnifyArc, Ungca32,
	{$IFNDEF EXTENSION}LObjects, {$ENDIF} EOFileIntf, ArchiversShared;

type
	TFileArray = array of TFile;

	TGCAArchive = class(TInterfacedObject, IFolder, IArchive)
	private
		FPath: TPath;
		FFiles: TFileArray;
		FViews: array of IObserver;
		FValid: Boolean;
		{IObservable}
		procedure Attach(const AView: IObserver);
		procedure Detach(const AView: IObserver);
		function HasClients: Boolean;
		{IFolder}
		function GetPath: TPath;
		function GetItems(AIndex: Integer): PFile;
		function GetCount: Integer;
		function GetValid: Boolean;
		function GetDisplayText: WideString;
		procedure Refresh;
		procedure Move(const NewPath: WideString);
		function GetParentPath: TPath;
		function IsSub(const AParent: TPath): Boolean;
		procedure StopSurveillance;
		procedure ResumeSurveillance(Wait: Boolean);
		function GetInfoText(AFile: PFile): WideString;
		procedure UnpackMemory(AFile: PFile; var Data: String);
		procedure Unpack(AFile: PFile; const DestDir: WideString);
		function GetPossibilities: TFolderPossibilities;
		function ConcatenatePath(AFile: PFile): TPath;
	protected
		procedure Notify;
		procedure DoRefresh;
	public
		constructor Create(const APath: TPath);
	end;

	TGCAArchiver = class(TInterfacedObject, IExtension, IExtensionInfo, IArchiver)
	public
		{IExtension}
		procedure Initialize;
		procedure Finalize;
		{IExtensionInfo}
		function GetName: WideString;
		procedure Dialog;
		{IArchiver}
		function Enumerate(Proc: CEachArchiveKind; Context: Pointer): WideString;
		function Open(const APath: TPath): IFolder;
		procedure Pack(const Archive, SourceDir: WideString; const Sources: array of WideString);
		procedure Unpack(const Archive, DestDir: WideString);
	end;

implementation

function ExtractArchiveDir(const S: WideString): WideString;
begin
	Result := S;
	Delete(Result, Max(1, WideLastDelimiter('/', Result)), MaxInt);
end;

{ TGCAArchive }

procedure TGCAArchive.Attach(const AView: IObserver);
begin
	SetLength(FViews, Length(FViews) + 1);
	FViews[High(FViews)] := AView;
end;

function TGCAArchive.ConcatenatePath(AFile: PFile): TPath;
var
	S: WideString;
begin
	Result := EmptyPath;
	
	Result.Directory := FPath.Directory;
	Result.Archive := FPath.Archive;

	with AFile^ do SetString(S, Name, NameLength);

	if Result.Archive <> '' then
		Result.Archive := Result.Archive + '/';

	Result.Archive := Result.Archive + S;
end;

constructor TGCAArchive.Create(const APath: TPath);
begin
	inherited Create;
	FPath := APath;

	SetCursorWait;
	DoRefresh;
	ResetCursor;
end;

procedure TGCAArchive.Detach(const AView: IObserver);
var
	I, J: Integer;
begin
	for I := High(FViews) downto Low(FViews) do
		if FViews[I] = AView then
		begin
			for J := I + 1 to High(FViews) do
				FViews[J - 1] := FViews[J];

			SetLength(FViews, Length(FViews) - 1);
			Break;
		end;
end;

procedure TGCAArchive.DoRefresh;
label
	Next;
const
	BufferFirst = 1024;
	BufferStep = 256;
var
	Info: TINDIVIDUALINFO;
	Sub, I, Count: Integer;
	H: THandle;
	FileName: WideString;
	PathInArc: string;
	F: TFile;
	PFileName: PChar;
begin
	FValid := False;

	Count := 0;
	SetLength(FFiles, BufferFirst);

	H := UnGCAOpenArchive(0, PChar(string(FPath.Directory)),
					M_REGARDLESS_INIT_FILE or M_NOT_USE_DRIVE_LETTER or
					M_CHECK_FILENAME_ONLY or M_ERROR_MESSAGE_ON or M_NO_BACKGROUND_MODE);
	if H = 0 then	Exit;

	PathInArc := '';
	if FPath.Archive <> '' then
	begin
		PathInArc := StringReplace(FPath.Archive, '/', '\', [rfReplaceAll]) + '\';
	end;

	if UnGCAFindFirst(H, '*', Info) = 0 then
		repeat
			PFileName := Info.szFileName;

			if StrLIComp(PFileName, PChar(PathInArc), Length(PathInArc)) <> 0 then
				Continue;

			Inc(PFileName, Length(PathInArc));
			if PFileName^ = #0 then Continue;

			FileName := StringReplace(PFileName, '\', '/', [rfReplaceAll]);
			if FileName[Length(FileName)] = '/' then
				Delete(FileName, Length(FileName), 1);

			Sub := Pos('/', FileName);
			if Sub > 0 then
			begin
				FileName := Copy(FileName, 1, Sub - 1);

				for I := Count - 1 downto 0 do
					if lstrcmpW(FFiles[I].Name, PWideChar(FileName)) = 0 then
						goto Next;

				F.Kind := fkFile;

				lstrcpyW(F.Name, PWideChar(FileName));
				F.NameLength := Length(FileName);
				F.DisplayName := F.Name;
				F.DisplayNameLength := F.NameLength;
				F.ShortNameLength := 0;

				F.CreationTime.dwLowDateTime := 0;
				F.CreationTime.dwHighDateTime := 0;
				F.LastAccessTime.dwLowDateTime := 0;
				F.LastAccessTime.dwHighDateTime := 0;
				F.LastWriteTime.dwLowDateTime := 0;
				F.LastWriteTime.dwHighDateTime := 0;

				F.Size := 0;
				F.Attributes := faDirectory;

				F.Operatable := [foEnter];
			end
			else
			begin
				F.Kind := fkFile;

				lstrcpyW(F.Name, PWideChar(FileName));
				F.NameLength := Length(FileName);
				F.DisplayName := F.Name;
				F.DisplayNameLength := F.NameLength;
				F.ShortNameLength := 0;

				F.CreationTime.dwLowDateTime := 0;
				F.CreationTime.dwHighDateTime := 0;
				F.LastAccessTime.dwLowDateTime := 0;
				F.LastAccessTime.dwHighDateTime := 0;
				UnGCAGetWriteTimeEx(H, F.LastWriteTime);

				F.Size := Info.dwOriginalSize;
				F.Attributes := UnGCAGetAttribute(H);

				F.Operatable := [];
				if F.Attributes and faDirectory <> 0 then
					Include(F.Operatable, foEnter);
			end;
			F.Number := High(Word);

			{Add}
			Inc(Count);
			if Count > Length(FFiles) then
				SetLength(FFiles, Length(FFiles) + BufferStep);
			FFiles[Count - 1] := F;
Next:
		until UnGCAFindNext(H, Info) <> 0;
	UnGCACloseArchive(H);

	SetLength(FFiles, Count);
	FValid := True;
end;

function TGCAArchive.GetCount: Integer;
begin
	Result := Length(FFiles)
end;

function TGCAArchive.GetDisplayText: WideString;
begin
	Result := FPath.Directory;
	if FPath.Archive <> '' then
		Result := Result + '/' + FPath.Archive
end;

function TGCAArchive.GetInfoText(AFile: PFile): WideString;
begin
	Result := 'in GCA'
end;

function TGCAArchive.GetItems(AIndex: Integer): PFile;
begin
	Result := @FFiles[AIndex]
end;

function TGCAArchive.GetParentPath: TPath;
begin
	Result := FPath;

	if Result.Archive = '' then
		Result.Directory := WideExtractFileDir(Result.Directory)
	else
		Result.Archive := ExtractArchiveDir(Result.Archive)
end;

function TGCAArchive.GetPath: TPath;
begin
	Result := FPath
end;

function TGCAArchive.GetPossibilities: TFolderPossibilities;
begin
	Result := [fpSort, fpHasParent]
end;

function TGCAArchive.GetValid: Boolean;
begin
	Result := FValid
end;

function TGCAArchive.HasClients: Boolean;
begin
	Result := FViews <> nil
end;

function TGCAArchive.IsSub(const AParent: TPath): Boolean;
var
	ParentDir, ParentArc: WideString;
	ParentNameSpace: INameSpace;
begin
	if AParent.ItemIDList = nil then
	begin
		if AParent.Directory = FPath.Directory then
		begin
			ParentArc := AParent.Archive + '/';
			Result := WideCompareFileName(ParentArc, Copy(FPath.Archive + '/', 1, Length(ParentArc))) = 0
		end
		else
		begin
			ParentDir := WideIncludeTrailingPathDelimiter(AParent.Directory);
			Result := WideCompareFileName(ParentDir, Copy(FPath.Directory + '\', 1, Length(ParentDir))) = 0
		end
	end
	else
	begin
		{fXNgbv}CRs[^}
		ParentNameSpace := (GetApplication as IFolderManager).GetFolder(AParent) as INameSpace;
		Result := ParentNameSpace.AtDesktop or ParentNameSpace.AtDrives;
	end;
end;

procedure TGCAArchive.Move(const NewPath: WideString);
begin
	FPath.Directory := NewPath
end;

procedure TGCAArchive.Notify;
var
	I: Integer;
	C: Pointer;
begin
(*
	BeginNotification;
	try
*)
		{EĂvȂ悤Ɂc}
		I := Low(FViews);
		while I <= High(FViews) do
		begin
			C := Pointer(FViews[I]);
			FViews[I].Observe;
			if C = Pointer(FViews[I]) then
				Inc(I);
		end;
(*
	finally
		EndNotification
	end
*)
end;

procedure TGCAArchive.Refresh;
var
	OldFiles: TFileArray;
begin
	if FValid then
	begin
		SetCursorWait;

		OldFiles := FFiles; {View ܂QƂĂ\邽߉Ȃ}
		try
			DoRefresh;
		except
			FValid := False;
		end;
		ResetCursor;

		Notify; {ʒm}
	end;

	OldFiles := nil; {sage...}
end;

procedure TGCAArchive.ResumeSurveillance;
begin
end;

procedure TGCAArchive.StopSurveillance;
begin
end;

procedure TGCAArchive.Unpack(AFile: PFile; const DestDir: WideString);
var
	DestPath, F, Full: WideString;
	Cmd: string;
	Buffer: array[0..4095] of Char;
begin
	DestPath := WideIncludeTrailingPathDelimiter(DestDir);

	with AFile^ do SetString(F, Name, NameLength);
	if FPath.Archive <> '' then F := FPath.Archive + '/' + F;

	Full := DestPath + F;

	Cmd := 'ex "' + FPath.Directory + '" "' + DestPath + '" "' + StringReplace(F, '/', '\', [rfReplaceAll]) + '"';
	if UnGCA(0, PChar(Cmd), Buffer, High(Buffer)) <> 0 then
		raise EUnGCAError.Create(Buffer);
end;

procedure TGCAArchive.UnpackMemory(AFile: PFile; var Data: String);
var
	Size: DWORD;
	Cmd: string;
	F: WideString;
	E: Integer;
begin
	Size := AFile^.Size;
	SetLength(Data, Size);
	with AFile^ do SetString(F, Name, NameLength);
	if FPath.Archive <> '' then F := FPath.Archive + '/' + F;
	Cmd := 'ex "' + FPath.Directory + '" .\ "' + StringReplace(F, '/', '\', [rfReplaceAll]) + '"';
	E := UngcaExtractMem(0, PChar(Cmd), Pointer(Data), Size, nil, nil, nil);
	if E <> 0 then
		raise EUnGcaError.Create('UngcaExtractMem: ' + IntToStr(E))
end;

{ TGCAArchiver }

procedure TGCAArchiver.Dialog;
var
	Buffer: array[0..1023] of Char;
begin
	try
		UngcaConfigDialog(0, Buffer, UNPACK_CONFIG_MODE)
	except
		MessageBoxW(0, 'Ungca32.dll܂', 'GCA', MB_ICONERROR)
	end
end;

function TGCAArchiver.Enumerate(Proc: CEachArchiveKind; Context: Pointer): WideString;
begin
	if Proc('.gca', Context) then
		Result := '.lzh'
	else
		Result := ''
end;

procedure TGCAArchiver.Finalize;
begin
end;

function TGCAArchiver.GetName: WideString;
begin
	Result := 'Ungca32.dll'
end;

procedure TGCAArchiver.Initialize;
begin
end;

function TGCAArchiver.Open(const APath: TPath): IFolder;
begin
	if not UngcaEnabled then
		raise EUngcaError.CreateFmt(SFaultLoad, [Ungca32Dll]);

	Result := TGCAArchive.Create(APath)
end;

procedure TGCAArchiver.Pack(const Archive, SourceDir: WideString; const Sources: array of WideString);
begin
	raise EUnGcaError.Create('GCAkɂ͑ΉĂ܂.');
end;

procedure TGCAArchiver.Unpack(const Archive, DestDir: WideString);
var
	Cmd: String;
	Buffer: array[0..4095] of Char;
begin
	Cmd := 'e "' + Archive + '" "' + IncludeTrailingPathDelimiter(DestDir) + '"';

	if UnGCA(0, PChar(Cmd), Buffer, High(Buffer)) <> 0 then
		raise EUnGcaError.Create(Buffer);
end;

procedure Register;
begin
	(GetApplication as IExtensionManager).RegisterExtension(TGCAArchiver.Create)
end;

initialization

	if GetInterfaceVersion = EOFileIntf.InterfaceVersion then
		Register;

end.
