unit ZIPModule;

interface

uses
	Windows, SysUtils, Math,
	UnifyArc, SevenZip32,
	{$IFNDEF EXTENSION}LObjects, {$ENDIF} EOFileIntf, ArchiversShared;

type
	TFileArray = array of TFile;

	TArchiveType = (atInvalid, atZip, at7z);

	TZIPArchive = class(TInterfacedObject, IFolder, IArchive)
	private
		FPath: TPath;
		FFiles: TFileArray;
		FViews: array of IObserver;
		FValid: Boolean;
		FType: TArchiveType;
		{IObservable}
		procedure Attach(const AView: IObserver);
		procedure Detach(const AView: IObserver);
		function HasClients: Boolean;
		{IFolder}
		function GetPath: TPath;
		function GetItems(AIndex: Integer): PFile;
		function GetCount: Integer;
		function GetValid: Boolean;
		function GetDisplayText: WideString;
		procedure Refresh;
		procedure Move(const NewPath: WideString);
		function GetParentPath: TPath;
		function IsSub(const AParent: TPath): Boolean;
		procedure StopSurveillance;
		procedure ResumeSurveillance(Wait: Boolean);
		function GetInfoText(AFile: PFile): WideString;
		procedure UnpackMemory(AFile: PFile; var Data: String);
		procedure Unpack(AFile: PFile; const DestDir: WideString);
		function GetPossibilities: TFolderPossibilities;
		function ConcatenatePath(AFile: PFile): TPath;
	protected
		procedure Notify;
		procedure DoRefresh;
	public
		constructor Create(const APath: TPath);
	end;

	TZIPArchiver = class(TInterfacedObject, IExtension, IExtensionInfo, IArchiver)
	public
		{IExtension}
		procedure Initialize;
		procedure Finalize;
		{IExtensionInfo}
		function GetName: WideString;
		procedure Dialog;
		{IArchiver}
		function Enumerate(Proc: CEachArchiveKind; Context: Pointer): WideString;
		function Open(const APath: TPath): IFolder;
		procedure Pack(const Archive, SourceDir: WideString; const Sources: array of WideString);
		procedure Unpack(const Archive, DestDir: WideString);
	end;

implementation

function MakeResponseFile(const BasePath: string; const Files: array of WideString): TFileName;
var
	H: THandle;
	F, Line: string;
	I: Integer;
begin
	Result := TemporaryPath + '$$$$$$$$.lst';

	H := FileCreate(Result);
	if H = INVALID_HANDLE_VALUE then
		raise EInOutError.CreateFmt('X|Xt@C%s쐬ł܂ł.', [Result]);

	try
		for I := Low(Files) to High(Files) do
		begin
			F := Files[I];

			if DirectoryExists(BasePath + F) then
				Line := '"' + F + '\*"'#13#10
			else
				Line := '"' + F + '"'#13#10;

			FileWrite(H, PChar(Line)^, Length(Line));
		end;
	finally
		FileClose(H);
	end;
end;

function ExtractArchiveDir(const S: WideString): WideString;
begin
	Result := S;
	Delete(Result, Max(1, WideLastDelimiter('/', Result)), MaxInt);
end;

{ TZIPArchive }

procedure TZIPArchive.Attach(const AView: IObserver);
begin
	SetLength(FViews, Length(FViews) + 1);
	FViews[High(FViews)] := AView;
end;

function TZIPArchive.ConcatenatePath(AFile: PFile): TPath;
var
	S: WideString;
begin
	Result := EmptyPath;

	Result.Directory := FPath.Directory;
	Result.Archive := FPath.Archive;

	with AFile^ do SetString(S, Name, NameLength);

	if Result.Archive <> '' then
		Result.Archive := Result.Archive + '/';

	Result.Archive := Result.Archive + S;
end;

constructor TZIPArchive.Create(const APath: TPath);
begin
	inherited Create;
	FPath := APath;

	SetCursorWait;
	DoRefresh;
	ResetCursor;
end;

procedure TZIPArchive.Detach(const AView: IObserver);
var
	I, J: Integer;
begin
	for I := High(FViews) downto Low(FViews) do
		if FViews[I] = AView then
		begin
			for J := I + 1 to High(FViews) do
				FViews[J - 1] := FViews[J];

			SetLength(FViews, Length(FViews) - 1);
			Break;
		end;
end;

procedure TZIPArchive.DoRefresh;
label
	Next;
const
	BufferFirst = 1024;
	BufferStep = 256;
var
	Info: TINDIVIDUALINFO;
	Count, I: Integer;
	H: THandle;
	ArcNameA, PathInArc: string;
	F: TFile;
	PFileName: PChar;
	Overwrite: Boolean;
	Sub: PWideChar;
begin
	FValid := False;

	Count := 0;
	SetLength(FFiles, BufferFirst);

	ArcNameA := string(FPath.Directory);

	case SevenZipGetArchiveType(PChar(ArcNameA)) of
	1:   FType := atZip;
	2:   FType := at7z;
	else FType := atInvalid;
	end;

	H := SevenZipOpenArchive(0, PChar(ArcNameA),
					M_REGARDLESS_INIT_FILE or M_NOT_USE_DRIVE_LETTER or
					M_CHECK_FILENAME_ONLY or M_ERROR_MESSAGE_ON or M_NO_BACKGROUND_MODE);
	if H = 0 then	Exit;

	PathInArc := '';
	if FPath.Archive <> '' then
		PathInArc := StringReplace(FPath.Archive, '/', '\', [rfReplaceAll]) + '\';

	if SevenZipFindFirst(H, '*', Info) = 0 then
		repeat
			PFileName := Info.szFileName;

			if StrLIComp(PFileName, PChar(PathInArc), Length(PathInArc)) <> 0 then
				Continue;

			Inc(PFileName, Length(PathInArc));
			if PFileName^ = #0 then
				Continue;

			F.NameLength := MultiByteToWideChar(CP_ACP, 0, PFileName, -1, F.Name, High(F.Name)) - 1;
			F.Name[F.NameLength] := #0;
			for I := F.NameLength - 1 downto 0 do
				if F.Name[I] = '\' then F.Name[I] := '/';

			Sub := WideStrScan(F.Name, '/');
			if Sub = F.Name + F.NameLength - 1 then
			begin
				Dec(F.NameLength);
				F.Name[F.NameLength] := #0;
				Sub := nil;
			end;

			if Sub <> nil then
			begin
				F.NameLength := Sub - F.Name;
				F.Name[F.NameLength] := #0;

				for I := Count - 1 downto 0 do
					if (FFiles[I].NameLength = F.NameLength) and
						CompareMem(@FFiles[I].Name, @F.Name, F.NameLength * SizeOf(WideChar))
					then
						goto Next;

				F.Kind := fkFile;
				F.DisplayName := F.Name;
				F.DisplayNameLength := F.NameLength;
				F.ShortNameLength := 0;

				F.CreationTime.dwLowDateTime := 0;
				F.CreationTime.dwHighDateTime := 0;
				F.LastAccessTime.dwLowDateTime := 0;
				F.LastAccessTime.dwHighDateTime := 0;
				F.LastWriteTime.dwLowDateTime := 0;
				F.LastWriteTime.dwHighDateTime := 0;

				F.Size := 0;
				F.Number := High(Word);
				F.Attributes := faDirectory;

				F.Operatable := [foEnter];
			end
			else
			begin
				Overwrite := (Count > 0) and
					(FFiles[Count - 1].NameLength = F.NameLength) and
					CompareMem(@FFiles[Count - 1].Name, @F.Name, F.NameLength * SizeOf(WideChar));

				F.Kind := fkFile;
				F.DisplayName := F.Name;
				F.DisplayNameLength := F.NameLength;
				F.ShortNameLength := 0;

				F.CreationTime.dwLowDateTime := 0;
				F.CreationTime.dwHighDateTime := 0;
				F.LastAccessTime.dwLowDateTime := 0;
				F.LastAccessTime.dwHighDateTime := 0;
				if not SevenZipGetWriteTimeEx(H, F.LastWriteTime) then
				begin
					F.LastWriteTime := UnixUTCToFileTime(SevenZipGetWriteTime(H))
				end;

				F.Size := Info.dwOriginalSize;
				F.Number := High(Word);
				F.Attributes := SevenZipGetAttribute(H);

				F.Operatable := [];
				if F.Attributes and faDirectory <> 0 then
					Include(F.Operatable, foEnter);

				if Overwrite then
				begin
					FFiles[Count - 1] := F;
					Continue;
				end;
			end;

			{Add}
			Inc(Count);
			if Count > Length(FFiles) then
				SetLength(FFiles, Length(FFiles) + BufferStep);
			FFiles[Count - 1] := F;
Next:
		until SevenZipFindNext(H, Info) <> 0;
	SevenZipCloseArchive(H);

	SetLength(FFiles, Count);
	FValid := True;
end;

function TZIPArchive.GetCount: Integer;
begin
	Result := Length(FFiles)
end;

function TZIPArchive.GetDisplayText: WideString;
begin
	Result := FPath.Directory;
	if FPath.Archive <> '' then
		Result := Result + '/' + FPath.Archive
end;

function TZIPArchive.GetInfoText(AFile: PFile): WideString;
const
	T: array[TArchiveType] of WideString = ('in unidentified', 'in ZIP', 'in 7z');
begin
	Result := T[FType]
end;

function TZIPArchive.GetItems(AIndex: Integer): PFile;
begin
	Result := @FFiles[AIndex]
end;

function TZIPArchive.GetParentPath: TPath;
begin
	Result := FPath;

	if Result.Archive = '' then
		Result.Directory := WideExtractFileDir(Result.Directory)
	else
		Result.Archive := ExtractArchiveDir(Result.Archive)
end;

function TZIPArchive.GetPath: TPath;
begin
	Result := FPath
end;

function TZIPArchive.GetPossibilities: TFolderPossibilities;
begin
	Result := [fpSort, fpHasParent]
end;

function TZIPArchive.GetValid: Boolean;
begin
	Result := FValid
end;

function TZIPArchive.HasClients: Boolean;
begin
	Result := FViews <> nil
end;

function TZIPArchive.IsSub(const AParent: TPath): Boolean;
var
	ParentDir, ParentArc: WideString;
	ParentNameSpace: INameSpace;
begin
	if AParent.ItemIDList = nil then
	begin
		if AParent.Directory = FPath.Directory then
		begin
			ParentArc := AParent.Archive + '/';
			Result := WideCompareFileName(ParentArc, Copy(FPath.Archive + '/', 1, Length(ParentArc))) = 0
		end
		else
		begin
			ParentDir := WideIncludeTrailingPathDelimiter(AParent.Directory);
			Result := WideCompareFileName(ParentDir, Copy(FPath.Directory + '\', 1, Length(ParentDir))) = 0
		end
	end
	else
	begin
		{fXNgbv}CRs[^}
		ParentNameSpace := (GetApplication as IFolderManager).GetFolder(AParent) as INameSpace;
		Result := ParentNameSpace.AtDesktop or ParentNameSpace.AtDrives;
	end;
end;

procedure TZIPArchive.Move(const NewPath: WideString);
begin
	FPath.Directory := NewPath
end;

procedure TZIPArchive.Notify;
var
	I: Integer;
	C: Pointer;
begin
(*
	BeginNotification;
	try
*)
		{EĂvȂ悤Ɂc}
		I := Low(FViews);
		while I <= High(FViews) do
		begin
			C := Pointer(FViews[I]);
			FViews[I].Observe;
			if C = Pointer(FViews[I]) then
				Inc(I);
		end;
(*
	finally
		EndNotification
	end
*)
end;

procedure TZIPArchive.Refresh;
var
	OldFiles: TFileArray;
begin
	if FValid then
	begin
		SetCursorWait;

		OldFiles := FFiles; {View ܂QƂĂ\邽߉Ȃ}
		try
			DoRefresh;
		except
			FValid := False;
		end;
		ResetCursor;

		Notify; {ʒm}
	end;

	OldFiles := nil; {sage...}
end;

procedure TZIPArchive.ResumeSurveillance;
begin
end;

procedure TZIPArchive.StopSurveillance;
begin
end;

procedure TZIPArchive.Unpack(AFile: PFile; const DestDir: WideString);
var
	DestPath: WideString;
	F, Full: WideString;
	Cmd: string;
	Buffer: array[0..4095] of Char;
begin
	DestPath := WideIncludeTrailingPathDelimiter(DestDir);

	with AFile^ do SetString(F, Name, NameLength);
	if FPath.Archive <> '' then F := FPath.Archive + '/' + F;

	Full := DestPath + F;

	{-aoa: ㏑}

	Cmd := 'e -hide -aoa "' + FPath.Directory + '" "' + DestPath + '" "' + F + '"';
	if SevenZip(0, PChar(Cmd), Buffer, High(Buffer)) <> 0 then
		raise ESevenZipError.Create(Buffer);
end;

procedure TZIPArchive.UnpackMemory(AFile: PFile; var Data: String);
var
	WorkS: WideString;
	F, Full: WideString;
	Cmd: string;
	Buffer: array[0..4095] of Char;
	H: THandle;
begin
	{7-zip32.dllExtractMemT|[gȂ̂tempɓWJ}
	WorkS := TemporaryPath;

	with AFile^ do SetString(F, Name, NameLength);
	Full := F;
	if FPath.Archive <> '' then F := FPath.Archive + '/' + F;

	{-aoa: ㏑}

	Cmd := 'e -aoa "' + FPath.Directory + '" "' + WorkS + '" "' + F + '"';
	if SevenZip(0, PChar(Cmd), Buffer, High(Buffer)) <> 0 then
		raise ESevenZipError.Create(Buffer);

	Full := WorkS + Full;

	H := FileOpen(Full, fmOpenRead or fmShareDenyWrite);
	if H = INVALID_HANDLE_VALUE then RaiseLastOSError;
	SetLength(Data, GetFileSize(H, nil));
	FileRead(H, Pointer(Data)^, Length(Data));
	FileClose(H);

	DeleteFileW(PWideChar(Full));
end;

{ TZIPArchiver }

procedure TZIPArchiver.Dialog;
var
	Buffer: array[0..1023] of Char;
begin
	try
		SevenZipConfigDialog(0, Buffer, UNPACK_CONFIG_MODE)
	except
		MessageBoxW(0, '7-zip32.dll܂', 'ZIP', MB_ICONERROR)
	end
end;

function TZIPArchiver.Enumerate(Proc: CEachArchiveKind; Context: Pointer): WideString;
begin
	if Proc('.zip', Context) then
		Result := '.zip'
	else if Proc('.7z', Context) then
		Result := '.7z'
	else
		Result := ''
end;

procedure TZIPArchiver.Finalize;
begin
end;

function TZIPArchiver.GetName: WideString;
begin
	Result := '7-zip32.dll'
end;

procedure TZIPArchiver.Initialize;
begin
end;

function TZIPArchiver.Open(const APath: TPath): IFolder;
begin
	if not SevenZipEnabled then
		raise ESevenZipError.CreateFmt(SFaultLoad, [SevenZip32Dll]);

	Result := TZIPArchive.Create(APath)
end;

procedure TZIPArchiver.Pack(const Archive, SourceDir: WideString;	const Sources: array of WideString);
const
	FormatOptions: array[Boolean] of string = ('-tzip ', '-t7z ');
var
	ArchiveA, SourcePath, Cmd, ResponseFile: string;
	Buffer: array[0..4095] of Char;
begin
	try
		ArchiveA := Archive;
		SourcePath := IncludeTrailingPathDelimiter(SourceDir);

		{R}h}
		Cmd := 'a -r0 ' + FormatOptions[SameText(ExtractFileExt(ArchiveA), '.7z')];

		{ɂƊ{pX}
		Cmd := Cmd + '"' + ArchiveA +'" "' + SourcePath + '"';

		{Ώ}
		if Length(Sources) = 1 then
		begin
			if DirectoryExists(SourcePath + Sources[0]) then
				Cmd := Cmd + ' "' + Sources[0] +'\*"'
			else
				Cmd := Cmd + ' "' + Sources[0] +'"'
		end
		else
		begin
			ResponseFile := MakeResponseFile(SourcePath, Sources);
			Cmd := Cmd + ' @' + ExtractShortPathName(ResponseFile)
		end;

		if SevenZip(0, PChar(Cmd), Buffer, High(Buffer)) <> 0 then
			raise ESevenZipError.Create(Buffer);
	finally
		if ResponseFile <> '' then Windows.DeleteFile(PChar(ResponseFile));
	end
end;

procedure TZIPArchiver.Unpack(const Archive, DestDir: WideString);
var
	Cmd: String;
	Buffer: array[0..4095] of Char;
begin
	Cmd := 'x "' + Archive + '" "' + IncludeTrailingPathDelimiter(DestDir) + '"';

	if SevenZip(0, PChar(Cmd), Buffer, High(Buffer)) <> 0 then
		raise ESevenZipError.Create(Buffer);
end;

procedure Register;
begin
	(GetApplication as IExtensionManager).RegisterExtension(TZIPArchiver.Create)
end;

initialization

	if GetInterfaceVersion = EOFileIntf.InterfaceVersion then
		Register;

end.
